# Copyright 2009-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import functools

from nose.tools import istest
from reahl.tofu import  Fixture, test, scenario
from reahl.stubble import EmptyStub, stubclass
from reahl.tofu import vassert, expected, NoException

from reahl.component.context import ExecutionContext
from reahl.component.exceptions import ProgrammerError, IsInstance, IsCallable, IncorrectArgumentError
from reahl.component.modelinterface import Field, FieldIndex, ReahlFields, exposed, Event,\
                             EmailField, PasswordField, BooleanField, IntegerField, \
                             DateField, DateConstraint, \
                             ValidationConstraint, RequiredConstraint, MinLengthConstraint, \
                             MaxLengthConstraint, PatternConstraint, AllowedValuesConstraint, \
                             EqualToConstraint, RemoteConstraint, IntegerConstraint, \
                             MaxValueConstraint, MinValueConstraint, Action, AdaptedMethod, secured,\
                             AccessRightsConstraint, Choice, ChoiceGroup, ChoiceField, MultiChoiceConstraint, MultiChoiceField,\
                             FileField, SingleFileConstraint, UploadedFile, FileSizeConstraint, \
                             MimeTypeConstraint, MaxFilesConstraint, SmallerThanConstraint, GreaterThanConstraint
from reahl.component.exceptions import AccessRestricted

class FieldMixin(object):
    def new_model_object(self):
        obj = EmptyStub()
        obj.an_attribute = u'field value'
        return obj
    
    def new_field(self, cls=Field, name=u'an_attribute', label=u'the label'):
        field = cls(label=label)
        field.bind(name, self.model_object)
        return field


class FieldFixture(Fixture, FieldMixin):
    pass

@istest
class FieldTests(object):
    @test(FieldFixture)
    def marshalling(self, fixture):
        """A Field marshalls a string representation of some input to a python object which is then set as
           an attribute on the model object to which the Field is bound to. (A Field also unmarshalls such an attribute)."""
        
        # How the marshalling is programmed
        class MyField(Field):
            def parse_input(self, unparsed_input):
                if unparsed_input == u'one': return 1
                return 0
        
            def as_input(self):
                if self.get_model_value() == 1: return u'one'
                return u'zero'

        field = fixture.new_field(cls=MyField)

        # From string to python object
        field.from_input(u'one')
        vassert( fixture.model_object.an_attribute == 1 )

        # From python object back to string
        fixture.model_object.an_attribute = 0
        vassert( field.as_input() == u'zero' )


    @test(FieldFixture)
    def validation(self, fixture):
        """Field input is validated by means of Constraints added to the field. Such validation 
           happens in two places: validation of the string input, and validation based on the 
           parsed python object."""
        
        # For testing, we just use a simple one-way marshalling field from string to int
        class MyField(Field):
            def parse_input(self, unparsed_input):
                return int(unparsed_input)

        field = fixture.new_field(cls=MyField)

        # How a validation_constraint is coded that checks the raw string input
        class StringLevelConstraint(ValidationConstraint):
            name = u'stringlevel'
            def validate_input(self, unparsed_input):
                if unparsed_input not in (u'1', u'2'):
                    raise self
    
        # How a validation_constraint is coded that checks the final python object
        class PythonLevelConstraint(ValidationConstraint):
            name = u'pythonlevel'
            def validate_parsed_value(self, parsed_value):
                if parsed_value > 1:
                    raise self
        
        # How constraints are added to the field
        field.add_validation_constraint(StringLevelConstraint(u''))
        field.add_validation_constraint(PythonLevelConstraint(u''))
        
        # Input passing both constraints work as expected
        with expected(NoException):
            field.from_input(u'1')
            vassert( fixture.model_object.an_attribute == 1 )

        # Input failing either validation_constraint result in the appropriate validation_constraint being raised
        with expected(StringLevelConstraint):
            field.from_input(u'unparsable string')
        with expected(PythonLevelConstraint):
            field.from_input(u'2')

        # If multiple constraints fail, the first one (in order added to the field) is the one reported
        field.add_validation_constraint(AllowedValuesConstraint([]))
        with expected(StringLevelConstraint):
            field.from_input(u'unparsable string')

    @test(FieldFixture)
    def field_metadata(self, fixture):
        """Fields provide metadata about their contents"""
        field = Field(default=2, required=True, label='A field')
        field.bind(u'fieldname', fixture.model_object)
        
        vassert( field.label == u'A field' )
        vassert( field.required )
        vassert( field.variable_name == u'fieldname' )  
        vassert( field.get_model_value() == 2 )

    @test(FieldFixture)
    def constraint_message(self, fixture):
        """Each validation_constraint has a customisable error message which can make use of the label supplied by 
           the Field it is added to, or any other attribute of the ValidationConstraint itself."""
           
        class MyConstraint(ValidationConstraint):
            name = u'myconstraint'
            one_attr = u'boerpampoene'
            another_attr = 20091124
        
        # The message template is given when constructed
        validation_constraint = MyConstraint(u'Bob âte $missing_attr $one_attr on $another_attr $label')
        fixture.field.add_validation_constraint(validation_constraint)

        # The validation_constraint gets its label from its field
        vassert( validation_constraint.label == fixture.field.label )

        # How the template is filled with attributes (and missing ones)
        expected = u'Bob âte $missing_attr boerpampoene on 20091124 the label'
        vassert( validation_constraint.message == expected )

    @test(Fixture)
    def meaning_of_required(self, fixture):
        """Making a Field required means adding a RequiredConstraint to it"""
        # Case: construction with default value for required
        field = Field()
        vassert( not field.required )       

        # Case: construction with required=True
        field = Field(required=True)
        vassert( field.required )
        vassert( field.get_validation_constraint_named(RequiredConstraint.name) )      
    

    @test(Fixture)
    def getting_modified_copy(self, fixture):
        """It is possible to get a modified copy of an existing field if you want to link it with 
           different constraints on a different input"""
        other_constraint = ValidationConstraint(u'Other error')
        other_constraint.name = u'other'
        field = Field()
        field.add_validation_constraint(other_constraint)

        # Getting a copy
        new_field = field.copy()
        vassert( new_field is not field )
        vassert( new_field.name == field.name )
        vassert( new_field.storage_object == field.storage_object )
        vassert( new_field.default == field.default )
        vassert( new_field.label == field.label )
        copied_other_constraint = new_field.get_validation_constraint_named(other_constraint.name)
        vassert( copied_other_constraint.field is new_field )
        new_validation_constraints = [i.__class__ for i in new_field.validation_constraints]
        old_validation_constraints = [i.__class__ for i in field.validation_constraints]
        vassert( new_validation_constraints == old_validation_constraints )
        vassert( new_field.validation_constraints != field.validation_constraints )
        vassert( new_field.validation_constraints is not field.validation_constraints )

        vassert( new_field.access_rights is not field.access_rights )
        vassert( new_field.access_rights.readable is field.access_rights.readable )
        vassert( new_field.access_rights.writable is field.access_rights.writable )

        # Getting a required copy
        vassert( not field.required )
        required_field = field.as_required(required_message=u'new required message')
        vassert( required_field.required )
        required = required_field.get_validation_constraint_named(RequiredConstraint.name)
        vassert( required.error_message.template == u'new required message' )

        # Getting copy that's not required
        field.make_required(u'')
        vassert( field.required )
        optional_field = field.as_optional()
        vassert( not optional_field.required )

        # Getting copy with a ValidationConstraint of certain type removed
        vassert( field.required )
        more_lax_field = field.as_without_validation_constraint(RequiredConstraint)
        vassert( not more_lax_field.required )
        
        # Getting copy with a new ValidationConstraint added
        field.make_optional()
        vassert( not field.required )
        more_strict_field = field.as_with_validation_constraint(RequiredConstraint())
        vassert( more_strict_field.required )
        

    @test(Fixture)
    def helpers_for_fields(self, fixture):
        """The @exposed decorator makes it simpler to bind Fields to an object."""
        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.field1 = IntegerField()
                fields.field2 = BooleanField()

        model_object = ModelObject()

        vassert( model_object.fields is model_object.fields )
        vassert( model_object.fields.field1.bound_to is model_object )
        vassert( model_object.fields.field2.bound_to is model_object )

    @test(Fixture)
    def helpers_for_fields2(self, fixture):
        """The ReahlFields class is an alternative way to make it simpler to bind Fields to an object.
           This reads a bit nicer, BUT does not currently play well with internationalising strings."""
        class ModelObject(object):
            fields = ReahlFields()
            fields.field1 = IntegerField()
            fields.field2 = BooleanField()

        class InheritingModelObject(ModelObject):
            fields = ReahlFields()
            fields.field3 = IntegerField()

        model_object = ModelObject()

        vassert( model_object.fields is model_object.fields )
        vassert( model_object.fields.field1.bound_to is model_object )
        vassert( model_object.fields.field2.bound_to is model_object )

        inheriting_object = InheritingModelObject()
        
        vassert( inheriting_object.fields is not model_object.fields )
        vassert( inheriting_object.fields is inheriting_object.fields )
        vassert( inheriting_object.fields.field1.bound_to is inheriting_object )
        vassert( inheriting_object.fields.field2.bound_to is inheriting_object )
        vassert( inheriting_object.fields.field3.bound_to is inheriting_object )

    @test(Fixture)
    def re_binding_behaviour_of_field_index(self, fixture):
        """FieldIndexes wont bind a field if it already is bound."""
        model_object1 = EmptyStub()
        model_object2 = EmptyStub()
        bound_field = Field()
        bound_field.bind(u'bound_field', model_object2)

        vassert( bound_field.is_bound )
        vassert( bound_field.bound_to is model_object2 )
        index = FieldIndex(model_object1)
        index.new_name_for_bound_field = bound_field
        vassert( index.new_name_for_bound_field.name is u'bound_field' )
        vassert( bound_field.bound_to is model_object2 )

    @test(Fixture)
    def helpers_for_events(self, fixture):
        """The @exposed decorator makes it simpler to collect Events on an object similar to how it is used for Fields."""
        class ModelObject(object):
            @exposed
            def events(self, fields):
                fields.event1 = Event()
                fields.event2 = Event()

        model_object = ModelObject()

        vassert( model_object.events is model_object.events )
        vassert( model_object.events.event1.bound_to is model_object.events.event1 )
        vassert( model_object.events.event2.bound_to is model_object.events.event2 )

        vassert( model_object.events.event1.name == u'event1' )

    @test(Fixture)
    def helpers_for_events2(self, fixture):
        """The @exposed decorator can be used to get FakeEvents at a class level, provided the valid Event names are specified."""
        class ModelObject(object):
            @exposed(u'event1')
            def events(self, fields):
                fields.event1 = Event()
                fields.event2 = Event()

        vassert( ModelObject.events.event1.name == u'event1' )

        with expected(AttributeError):
            ModelObject.events.nonevent

    @test(Fixture)
    def helpers_for_events3(self, fixture):
        """An Event has to be created for each of the names listed to the @exposed decorator, else an error is raised."""
        class ModelObject(object):
            @exposed(u'event1', u'forgotten')
            def events(self, fields):
                fields.event1 = Event()
                fields.event2 = Event()

        def check_exc(exc):
            vassert( str(exc).startswith(u'You promised to instantiate') )
        with expected(ProgrammerError, test=check_exc):
            ModelObject().events

    @test(Fixture)
    def events(self, fixture):
        """An Event defines a signal that can be sent to the system, with the intention to
           possibly trigger the execution of an Action by the system. Metadata, such as what
           a human might label the Event, is also specified."""
        
        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event(action=Action(self.do_something), label=u'human readable label')
                
            def do_something(self):
                self.something_done = True

        model_object = ModelObject()
        event = model_object.events.an_event.with_arguments()
        event.from_input(event.as_input())
        event.fire()

        vassert( model_object.something_done )
        vassert( model_object.events.an_event.label == u'human readable label' )

    @test(Fixture)
    def arguments_to_actions(self, fixture):
        """Arguments can be defined on an Event in order to be able to pass them to the linked Action 
           as args or kwargs as specified by the Action."""

        expected_arg = 123
        expected_kwarg = 45

        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event(one_argument=IntegerField(required=True),
                                        another_argument=IntegerField(),
                                        unused_argument=IntegerField(),
                                        action=Action(self.do_something,
                                                      [u'one_argument'],
                                                      dict(a_kwarg=u'another_argument')))
                
            def do_something(self, an_arg, a_kwarg=None):
                self.passed_an_arg = an_arg
                self.passed_a_kwarg = a_kwarg

        model_object = ModelObject()
        event = model_object.events.an_event.with_arguments(one_argument=123, another_argument=45, unused_argument=678)
        event.from_input(event.as_input())
        event.fire()

        vassert( model_object.passed_an_arg is expected_arg )
        vassert( model_object.passed_a_kwarg is expected_kwarg )


    @test(Fixture)
    def arguments_to_event(self, fixture):
        """Only Action objects can be sent as action= when creating an Event. The arguments passed to readable and writable
           should be callable objects with correct signature."""

        # action=
        with expected(IsInstance):
            Event(action=EmptyStub())

        def check_exc(expected_message, ex):
            message = str(ex).split(u':')[1][1:]
            vassert( message.startswith(expected_message) )

        # readable/writable are callable
        with expected(IsCallable, test=functools.partial(check_exc, u'readable should be a callable object')):
            Event(readable=EmptyStub())
        with expected(IsCallable, test=functools.partial(check_exc, u'writable should be a callable object')):
            Event(writable=EmptyStub())

        # readable/writable have correct signature
        def do_nothing(a, b, c, d): pass
        with expected(IncorrectArgumentError):
            Event(readable=do_nothing)
        with expected(IncorrectArgumentError):
            Event(writable=do_nothing)


    @test(Fixture)
    def event_security(self, fixture):
        """If an Event specifies an Action, the access controls of the Action are
           used for access to the Event as well."""
        class ModelObject(object):
            allow_read_flag = True
            allow_write_flag = True
            def allow_read(self):
                return self.allow_read_flag

            def allow_write(self):
                return self.allow_write_flag

            @secured( read_check=allow_read, write_check=allow_write )
            def do_something(self):
                pass

            @exposed
            def events(self, events):
                events.an_event = Event(action=Action(self.do_something))


        model_object = ModelObject()
        event = model_object.events.an_event.with_arguments()
        event.from_input(event.as_input())
        
        event.fire()
        
        vassert( event.can_read() )
        vassert( event.can_write() )

        model_object.allow_read_flag = False
        vassert( not event.can_read() )

        model_object.allow_write_flag = False
        vassert( not event.can_write() )

        with expected(ProgrammerError):
            event.fire()



    @test(Fixture)
    def event_security2(self, fixture):
        """If an Event does not specify an Action, then Actions can be passed for its readable and writable."""
        class ModelObject(object):
            allow_read_flag = True
            allow_write_flag = True
            def allow_read(self):
                return self.allow_read_flag

            def allow_write(self):
                return self.allow_write_flag

            @exposed
            def events(self, events):
                events.an_event = Event(readable=Action(self.allow_read),
                                        writable=Action(self.allow_write))

        model_object = ModelObject()
        event = model_object.events.an_event.with_arguments()

        vassert( event.can_read() )
        vassert( event.can_write() )

        model_object.allow_read_flag = False
        vassert( not event.can_read() )

        model_object.allow_write_flag = False
        vassert( not event.can_write() )

    @test(Fixture)
    def event_security_action_and_rw(self, fixture):
        """Supply either an action or a readable/writable to an Event, but not both."""

        def do_nothing(): pass

        with expected(ProgrammerError):
            Event(action=Action(do_nothing), readable=Action(do_nothing))
        with expected(ProgrammerError):
            Event(action=Action(do_nothing), writable=Action(do_nothing))

    @test(Fixture)
    def receiving_events(self, fixture):
        """An Event is said to have occurred if it received a querystring containing its arguments from user input.
           An Event can only be fired if it occurred."""

        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event(an_argument=IntegerField())

        model_object = ModelObject()
        event = model_object.events.an_event.with_arguments(an_argument=123)
        
        vassert( event.default == {u'an_argument': 123} )
        vassert( not hasattr(event, u'arguments') )
        vassert( not event.occurred )
        with expected(ProgrammerError):
            event.fire()
        
        event.from_input(u'?an_argument=123')
        vassert( event.arguments == {u'an_argument': 123} )
        vassert( event.occurred )
        with expected(NoException):
            event.fire()


    class Scenarios(Fixture):
        @scenario
        def both_disallowed(self):
            self.allow_read = False
            self.allow_write = False
            self.expected_occurred = False
            self.expected_exception = AccessRightsConstraint
        @scenario
        def read_allowed(self):
            self.allow_read = True
            self.allow_write = False
            self.expected_occurred = False
            self.expected_exception = AccessRightsConstraint
        @scenario
        def write_allowed(self):
            self.allow_read = False
            self.allow_write = True
            self.expected_occurred = False
            self.expected_exception = AccessRightsConstraint
        @scenario
        def both_allowed(self):
            self.allow_read = True
            self.allow_write = True
            self.expected_occurred = True
            self.expected_exception = NoException

    @test(Scenarios)
    def security_of_receiving_events(self, fixture):
        """An Event can only occur if BOTH its access restrictions are allowed."""

        class ModelObject(object):
            def allow_read(self):
                return fixture.allow_read
            def allow_write(self):
                return fixture.allow_write
            @exposed
            def events(self, events):
                events.an_event = Event(readable=Action(self.allow_read), writable=Action(self.allow_write))

        model_object = ModelObject()
        event = model_object.events.an_event
        
        vassert( not event.occurred )
        with expected(fixture.expected_exception):
            event.from_input(u'?')
        vassert( event.occurred is fixture.expected_occurred )



@istest
class SpecificConstraintTests(object):
    @test(Fixture)
    def required_constraint(self, fixture):
        selector = 'find me'
        required_constraint = RequiredConstraint(selector_expression=selector)
        
        #selector is returned as parameter
        vassert( required_constraint.parameters == selector )
        
        #case: no input 
        with expected(RequiredConstraint):
            required_constraint.validate_input(u'')
        with expected(RequiredConstraint):
            required_constraint.validate_input(None)
        #just spaces
        space=u' '                              
        with expected(RequiredConstraint):
            required_constraint.validate_input(space)
        with expected(RequiredConstraint):
            required_constraint.validate_input(space*56)
        
        #case: input 
        with expected(NoException):
            required_constraint.validate_input(u' something valid    ')
        with expected(NoException):
            required_constraint.validate_input(u'sômething else that_is_valid')
        with expected(NoException):
            required_constraint.validate_input(u'.')

    @test(Fixture)
    def min_length_constraint(self, fixture):
        min_required_length = 5
        min_length_constraint = MinLengthConstraint(min_length=min_required_length)
        
        #min length is returned as parameter
        vassert( min_length_constraint.parameters == u'%s' % min_required_length )
        
        #case: just enough characters, exact number of
        with expected(NoException):
            min_length_constraint.validate_input(u'⁵'*min_required_length)
        
        #case: more than enough
        with expected(NoException):
            min_length_constraint.validate_input(u'ê'*(min_required_length+1))
            
        #case: just not enough characters
        with expected(MinLengthConstraint):
            min_length_constraint.validate_input(u'ś'*(min_required_length-1))
        
        #case: empty string
        with expected(MinLengthConstraint):
            min_length_constraint.validate_input(u'')
                        

    @test(Fixture)
    def max_length_constraint(self, fixture):
        max_allowed_length = 5
        max_length_constraint = MaxLengthConstraint(max_length=max_allowed_length)
        
        #max length is returned as parameter
        vassert( max_length_constraint.parameters == u'%s' % max_allowed_length )
        
        #case: just enough characters, exact number of
        with expected(NoException):
            max_length_constraint.validate_input(u'⁵'*max_allowed_length)
        
        #case: less than max
        with expected(NoException):
            max_length_constraint.validate_input(u'⁵'*(max_allowed_length-1))
            
        #case: nothing
        with expected(NoException):
            max_length_constraint.validate_input(u'')

        #case: just too much
        with expected(MaxLengthConstraint):
            max_length_constraint.validate_input(u'ś'*(max_allowed_length+1))


    @test(Fixture)
    def pattern_constraint(self, fixture):
        allow_pattern = u'(ab)+'
        pattern_constraint = PatternConstraint(pattern=allow_pattern)
        
        #pattern is returned as parameter
        vassert( pattern_constraint.parameters == allow_pattern )
        
        #case: valid
        valid_input = 'ababab'
        with expected(NoException):
            pattern_constraint.validate_input(valid_input )
        
        #case invalid
        with expected(PatternConstraint):
            pattern_constraint.validate_input(u'aba')
        
        #case: faulty pattern
        faulty_pattern_constraint = PatternConstraint(pattern='??????')
        with expected(ProgrammerError):
            faulty_pattern_constraint.validate_input(valid_input)
        
        
    @test(Fixture)
    def allowed_values_constraint(self, fixture):
        allowed_values=[u'a',u'b']
        allowed_values_constraint = AllowedValuesConstraint(allowed_values=allowed_values)

        #case: valid input
        valid_input = allowed_values[1]
        with expected(NoException):
            allowed_values_constraint.validate_input(valid_input)

        #case: invalid input
        invalid_input = 'ba'
        assert invalid_input not in allowed_values
        with expected(AllowedValuesConstraint):
            allowed_values_constraint.validate_input(invalid_input)

    @test(Fixture)
    def equal_to_constraint(self, fixture):
        other_field = Field(label=u'other')
        equal_to_constraint = EqualToConstraint(other_field, u'$label, $other_label')
        other_field.set_user_input(u'should be equal to this string', ignore_validation=True)

        #case: valid input
        with expected(NoException):
            equal_to_constraint.validate_parsed_value(u'should be equal to this string' )

        #case: invalid input
        with expected(EqualToConstraint):
            equal_to_constraint.validate_parsed_value(u'this is not equal')

    @test(Fixture)
    def smaller_than_constraint(self, fixture):
        other_field = IntegerField(label=u'other')
        smaller_than_constraint = SmallerThanConstraint(other_field, u'$label, $other_label')
        other_field.set_user_input(u'5', ignore_validation=True)

        #case: valid input
        with expected(NoException):
            smaller_than_constraint.validate_parsed_value( 4 )

        #case: invalid input
        with expected(SmallerThanConstraint):
            smaller_than_constraint.validate_parsed_value( 5 )

    @test(Fixture)
    def greater_than_constraint(self, fixture):
        other_field = IntegerField(label=u'other')
        greater_than_constraint = GreaterThanConstraint(other_field, u'$label, $other_label')
        other_field.set_user_input(u'5', ignore_validation=True)

        #case: valid input
        with expected(NoException):
            greater_than_constraint.validate_parsed_value( 6 )

        #case: invalid input
        with expected(GreaterThanConstraint):
            greater_than_constraint.validate_parsed_value( 5 )


@istest
class SpecificFieldsTests(object):
    @test(Fixture)
    def email_validation(self, fixture):
        field = EmailField()
        
        # Case: invalid
        invalid_addresses = ['somethingWithoutTheAt', '@somethingThatBeginsWithTheAT',
                                      'somethingThatEndsWithTheAt@', '@' ,None]
        for address in invalid_addresses:
            with expected(PatternConstraint):
                field.set_user_input(address)
            vassert( field.validation_error is field.get_validation_constraint_named(u'pattern') )

        # Case: valid
        valid_addresses = ['someone@home.co.za', 'something@somewhere.com', 'j@j.ca']
        for address in valid_addresses:
            with expected(NoException):
                field.set_user_input(address, ignore_validation=True)
            vassert( not field.validation_error )

    @test(Fixture)
    def password_validation(self, fixture):
        field = PasswordField()
        
        # Case: invalid
        with expected(MinLengthConstraint):
            field.set_user_input(u'123')
        vassert( field.validation_error is field.get_validation_constraint_named(u'minlength') )
        with expected(MaxLengthConstraint):
            field.set_user_input(u'1'*21)
        vassert( field.validation_error is field.get_validation_constraint_named(u'maxlength') )

        # Case: valid
        with expected(NoException):
            field.set_user_input(u'my passwôrd')
        vassert( not field.validation_error )

    @test(Fixture)
    def password_access(self, fixture):
        """A PasswordField is world writable, but not readable by anyone."""

        field = PasswordField()
        field.bind(u'password_field', fixture)
        vassert( not field.can_read() )
        vassert( field.can_write() )

    @test(Fixture)
    def boolean_validation(self, fixture):
        obj = EmptyStub()
        field = BooleanField()
        field.bind('boolean_attribute', obj)
        
        # Case: invalid
        invalid_boolean_name = [u'negative', u'affirmative', u'+', u'-', None]
        for boolean_candidate in invalid_boolean_name:
            with expected(AllowedValuesConstraint):
                field.set_user_input(boolean_candidate)
            vassert( field.validation_error is field.get_validation_constraint_named(u'pattern') )

        # Case: valid
        field.from_input(u'on')
        vassert( obj.boolean_attribute is True )
        vassert( field.as_input() == u'on' )
        field.from_input(u'off')
        vassert( obj.boolean_attribute is False )
        vassert( field.as_input() == u'off' )

        # Case: required means True for BooleanField
        field = BooleanField(required=True)
        field.bind('boolean_attribute', obj)
        with expected(AllowedValuesConstraint):
            field.set_user_input(u'off')        
        vassert( field.validation_error is field.get_validation_constraint_named(u'pattern') )
        with expected(NoException):
            field.from_input(u'on')

    @test(Fixture)
    def boolean_i18n(self, fixture):
        @stubclass(ExecutionContext)
        class AfrikaansContext(ExecutionContext):
            @property
            def interface_locale(self):
                return u'af'
            
        with AfrikaansContext():
            #cs context.force_interface_locale = u'af'
            obj = EmptyStub()
            field = BooleanField()
            field.bind('boolean_attribute', obj)

            # Case: valid
            field.from_input(u'aan')
            vassert( obj.boolean_attribute is True )
            vassert( field.as_input() == u'aan' )
            field.from_input(u'af')
            vassert( obj.boolean_attribute is False )
            vassert( field.as_input() == u'af' )


    @test(Fixture)
    def integer_validation(self, fixture):
        field = IntegerField()

        # Case invalid
        with expected(IntegerConstraint):
            field.set_user_input(u'sdfdf')

        # Case valid
        with expected(NoException):
            field.set_user_input(u'3')

        # Case Max
        field = IntegerField(max_value=4)
        with expected(MaxValueConstraint):
            field.set_user_input(u'5')

        # Case Min
        field = IntegerField(min_value=4)
        with expected(MinValueConstraint):
            field.set_user_input(u'3')


    @test(Fixture)
    def integer_marshalling(self, fixture):
        field = IntegerField()
        obj = EmptyStub()
        field.bind('integer_value', obj)

        # From input
        field.from_input(u'5')
        vassert( obj.integer_value is 5 )

        # As input
        obj.integer_value = 6
        vassert( field.as_input() == u'6' )


    class ChoiceFixture(Fixture):
        def new_field(self, field_class=None):
            field_class = field_class or ChoiceField
            field = field_class(self.choices)
            field.bind('choice_value', self.model_object)
            return field

        def new_model_object(self):
            return EmptyStub()
            
        @scenario
        def plain_choices(self):
            self.all_choices = [Choice(1, IntegerField(label=u'One')), 
                                Choice(u'2', Field(label=u'Two'))]
            self.choices = self.all_choices
            self.groups = []
            self.valid_inputs = [u'1', u'2']
            self.input_to_value_map = {u'1': 1, u'2': u'2'}
            self.expected_validation_constraint = AllowedValuesConstraint

        @scenario
        def grouped_choices(self):
            self.all_choices = [Choice(1, IntegerField(label=u'One')), 
                                Choice(u'2', Field(label=u'Two'))]
            self.groups = [ChoiceGroup(u'', self.all_choices)]
            self.choices = self.groups
            self.valid_inputs = [u'1', u'2']
            self.input_to_value_map = {u'1': 1, u'2': u'2'}
            self.expected_validation_constraint = AllowedValuesConstraint

        @scenario
        def multi_choices(self):
            self.all_choices = [Choice(1, IntegerField(label=u'One')), 
                                Choice(2, IntegerField(label=u'Two')),
                                Choice(3, IntegerField(label=u'Three'))]
            self.groups = []
            self.choices = self.all_choices
            self.field = self.new_field(MultiChoiceField)

            self.valid_inputs = [(u'1',), [u'1', u'2']]
            self.input_to_value_map = {(u'1',): [1], (u'1', u'2'): [1,2]}
            self.expected_validation_constraint = MultiChoiceConstraint

    @test(ChoiceFixture)
    def choice_querying(self, fixture):
        """A ChoiceField maintains a list of the Choices and Groups it was constructed with."""
        field = fixture.field
        
        vassert( field.groups == fixture.groups )
        vassert( field.flattened_choices == fixture.all_choices )


    @test(ChoiceFixture)
    def choice_validation(self, fixture):
        """Input for a ChoiceField is valid only if it matches one of its Choices."""
        field = fixture.field

        # Case invalid
        with expected(fixture.expected_validation_constraint):
            field.set_user_input(u'sdfdf')

        # Case valid
        for i in fixture.valid_inputs:
            with expected(NoException):
                field.set_user_input(i)


    @test(ChoiceFixture)
    def choice_field_marshalling(self, fixture):
        """Input for a ChoiceField is marshalled differently, depending on the Field specified for the Choice matching the input."""
        field = fixture.field
        obj = fixture.model_object
        
        # From input
        for valid_input, value in fixture.input_to_value_map.items():
            field.from_input(valid_input)
            vassert( obj.choice_value == value )

        # As input
        for expected_input, value in fixture.input_to_value_map.items():
            obj.choice_value = value
            if isinstance(expected_input, tuple):
                expected_input = list(expected_input)
            vassert( field.as_input() == expected_input )


    @test(FieldFixture)
    def file_marshalling(self, fixture):
        """A FileField receives as input a list of UploadedFile objects.  Its marshalling job consists
           of merely changing such a list into a single value, or returning it as is depending on
           the setting of its allow_multiple flag.
        """

        field = FileField()
        obj = fixture.model_object
        field.bind('file_value', obj)

        files = [EmptyStub(), EmptyStub()]

        # Single file only
        field.from_input(files[:1])
        vassert( obj.file_value == files[0] )

        obj.file_value = files[0]
        vassert( field.as_input() == u'' )

        # Multiple files
        field = FileField(allow_multiple=True)
        field.bind('file_value', obj)

        field.from_input(files)
        vassert( obj.file_value == files )

        obj.file_value = files
        vassert( field.as_input() == u'' )

    @test(FieldFixture)
    def file_validation(self, fixture):
        """A FileField needs to check that the right number of files were submitted, depending on the
           setting of allow_multiple and/or required.
        """

        field = FileField()
        obj = fixture.model_object
        field.bind('file_value', obj)

        files = [EmptyStub(), EmptyStub()]

        # Single file only
        with expected(SingleFileConstraint):
            field.set_user_input(files)

        with expected(NoException):
            field.set_user_input(files[:1])

        # Single file that is required
        field = FileField(required=True)
        field.bind('file_value', obj)

        with expected(NoException):
            field.set_user_input(files[:1])
            
        with expected(RequiredConstraint):
            field.set_user_input([])

        # Multiple files
        field = FileField(allow_multiple=True)
        field.bind('file_value', obj)

        with expected(NoException):
            field.set_user_input(files)

    @test(FieldFixture)
    def file_validation_size(self, fixture):
        """A FileField can also limit the size if files uploaded.
        """

        field = FileField(allow_multiple=True, max_size_bytes=100)
        obj = fixture.model_object
        field.bind('file_value', obj)

        files = [UploadedFile(u'file1', EmptyStub(), u'', 100), UploadedFile(u'file2', EmptyStub(), u'', 50)]
        with expected(NoException):
            field.set_user_input(files)

        files = [UploadedFile(u'file1', EmptyStub(), u'', 100), UploadedFile(u'file2', EmptyStub(), u'', 200)]
        with expected(FileSizeConstraint):
            field.set_user_input(files)

    @test(FieldFixture)
    def file_validation_mime_type(self, fixture):
        """A FileField can also limit the mimetype of files allowed to be uploaded.
        """

        field = FileField(allow_multiple=True, accept=['text/*'])
        obj = fixture.model_object
        field.bind('file_value', obj)

        files = [UploadedFile(u'file1', EmptyStub(), u'text/html', 100), UploadedFile(u'file2', EmptyStub(), u'text/xml', 50)]
        with expected(NoException):
            field.set_user_input(files)

        files = [UploadedFile(u'file1', EmptyStub(), u'text/html', 100), UploadedFile(u'file2', EmptyStub(), u'application/java', 200)]
        with expected(MimeTypeConstraint):
            field.set_user_input(files)

    @test(FieldFixture)
    def file_validation_max_files(self, fixture):
        """A maximum can be placed upon the number of files that may be uploaded.
        """

        field = FileField(allow_multiple=True, max_files=1)
        obj = fixture.model_object
        field.bind('file_value', obj)

        files = [UploadedFile(u'file1', EmptyStub(), u'', 100)]
        with expected(NoException):
            field.set_user_input(files)

        files = [UploadedFile(u'file1', EmptyStub(), u'', 100), UploadedFile(u'file2', EmptyStub(), u'', 200)]
        with expected(MaxFilesConstraint):
            field.set_user_input(files)


    @test(FieldFixture)
    def date_marshalling(self, fixture):
        """A DateField marshalls human readable date representation to a datetime.date object.
           (It tolerates non-precise input.)
        """

        field = DateField()
        obj = fixture.model_object
        
        field.bind(u'date_value', obj)
        
        # From input
        for input_string in [u'10 November 2012', u'10/11/2012']:
            field.from_input(input_string)
            vassert( obj.date_value == datetime.date(2012, 11, 10) )

        # As input
        obj.date_value = datetime.date(2010, 11, 10)
        actual_output = field.as_input()
        vassert( actual_output == u'10 Nov 2010' )

    @test(FieldFixture)
    def date_validation(self, fixture):
        """A DateField can validate its input based on a min or max value and expects fuzzy but sensible input."""

        field = DateField()
        obj = fixture.model_object
        
        field.bind(u'date_value', obj)

        # Case invalid
        with expected(DateConstraint):
            field.set_user_input(u'sdfdf')

        # Case valid
        with expected(NoException):
            field.set_user_input(u'13 Dec')

        limit_date = datetime.date(2012, 11, 13)
        before_limit = u'12 Nov 2012'
        after_limit = u'14 Nov 2012'

        # Case Max
        field = DateField(max_value=limit_date)
        with expected(MaxValueConstraint):
            field.set_user_input(after_limit)

        # Case Min
        field = DateField(min_value=limit_date)
        with expected(MinValueConstraint):
            field.set_user_input(before_limit)






# Copyright 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Support for database schema migration."""

from pkg_resources import parse_version

from reahl.component.exceptions import ProgrammerError

class Migration(object):
    """Represents one logical change that can be made to an existing database schema.
    
       You should extend this class to build your own domain-specific database schema migrations. Set the
       `version` class attribute to a string containing the version of your component for which this Migration
       should be run when upgrading from a previous version.
       
       Never use code imported from your component in a Migration, since Migration code is kept around in
       future versions of a component and may be run to migrate a schema with different versions of your component.
    """
    version = None
    @classmethod
    def is_applicable(cls, current_schema_version, new_version):
        if not cls.version:
            raise ProgrammerError(u'Migration %s does not have a version set' % cls)
        return parse_version(cls.version) > parse_version(current_schema_version) and \
               parse_version(cls.version) <= parse_version(new_version)

    def run_phase(self, phase):
        if phase == 1:
            self.upgrade()
        else:
            self.upgrade_cleanup()

    def upgrade(self):
        """Override this method in a subclass in order to supply custom logic for changing the database schema. This
           method will be called for each of the applicable Migrations listed for all components, in reverse order of 
           dependency of components (less dependent components first).
        """
        pass
    def upgrade_cleanup(self):
        """Override this method in a subclass in order to supply custom logic that needs to run after the `.upgrade` method
           has been run for all Migrations of all components in an application. This
           method will be called for each of the applicable Migrations listed for all components, in normal order of 
           dependency of components (most dependent components first).
        """
        pass




# Copyright 2009-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The Reahl production commandline utility."""

import os
import re
import os.path
import sys
from optparse import Values, OptionParser
import shlex
from functools import wraps
import logging

from pkg_resources import DistributionNotFound

from reahl.component.dbutils import SystemControl
from reahl.component.shelltools import Command, ReahlCommandline
from reahl.component.context import ExecutionContext
from reahl.component.config import EntryPointClassList, Configuration, StoredConfiguration, MissingValue
from reahl.component.eggs import ReahlEgg
from reahl.component.decorators import memoized



class ProdShellConfig(Configuration):
    commands = EntryPointClassList(u'reahl.component.prodcommands', description=u'The commands (classes) available to the production commandline shell')


class ProductionCommand(Command):
    """Superclass used for all production shell commands."""
    options = []
    usage_args = '<config_directory>'
    
    def execute(self, options, args):
        self.create_context(args[0])
        self.directory = args[0]

    def verify_commandline(self, options, args):
        if not len(args) >= 1:
            self.parser.error(u'No config directory given')

    def create_context(self, config_directory):
        try:
            self.context = ExecutionContext.for_config_directory(config_directory)
        except DistributionNotFound, ex:
            ex.args = (u'%s (In development? Did you forget to do a "reahl setup -- develop -N"?)' % ex.args[0],)
            raise
        with self.context:
            self.context.system_control = SystemControl(self.context.config)

    @property
    def sys_control(self):
        return self.context.system_control
    
    @property
    def config(self):
        return self.context.config


class ListConfig(ProductionCommand):
    """Lists current configuration settings."""
    keyword = 'listconfig'
    options = ProductionCommand.options +\
               [('-v', '--values', dict(action='store_true', dest='print_values', help='prints the currently configured value')),
               ('-f', '--files', dict(action='store_true', dest='print_files', help='prints the filename where the setting should be defined')),
               ('-d', '--defaults', dict(action='store_true', dest='print_defaults', help='prints the default value')),
               ('-m', '--missing', dict(action='store_true', dest='print_missing_only', help='prints the missing values only')),
               ('-i', '--info', dict(action='store_true', dest='print_description', help='prints a description'))]

    def create_context(self, config_directory):
        self.context = ExecutionContext()

    def execute(self, options, args):
        super(ListConfig, self).execute(options, args)
        with self.context:
            print 'Listing config for %s' % self.directory
            config = StoredConfiguration(self.directory)
            config.configure(validate=False)
            for config_file, key, value, setting in config.list_all():
                to_print = '%-35s' % key
                if options.print_files:
                    to_print += '\t%s' % config_file
                if options.print_values:
                    to_print += '\t%s' % value
                if options.print_defaults:
                    if setting.defaulted:
                        message = setting.default
                        if setting.dangerous:
                            message += u' (DANGEROUS DEFAULT)'
                    elif setting.automatic:
                        message = u'AUTOMATIC'
                    else:
                        message = u'NO DEFAULT'
                    to_print += '\t%s' % message
                if options.print_description:
                    to_print += '\t%s' % setting.description

                if options.print_missing_only and not isinstance(value, MissingValue):
                    pass
                else:
                    print to_print


class CheckConfig(ProductionCommand):
    """Checks current configuration settings."""
    keyword = 'checkconfig'
    def execute(self, options, args):
        super(CheckConfig, self).execute(options, args)
        print 'Checking config in %s' % self.directory
        config = StoredConfiguration(self.directory)
        config.configure(validate=True)
        print 'Config parses OK'


class CreateDBUser(ProductionCommand):
    """Creates the database user."""
    keyword = 'createdbuser'
    def execute(self, options, args):
        super(CreateDBUser, self).execute(options, args)
        return self.sys_control.create_db_user()


class DropDBUser(ProductionCommand):
    """Drops the database user."""
    keyword = 'dropdbuser'
    def execute(self, options, args):
        super(DropDBUser, self).execute(options, args)
        return self.sys_control.drop_db_user()


class DropDB(ProductionCommand):
    """Drops the database."""
    keyword = 'dropdb'
    options = [('-y', '--yes', dict(action='store_true', dest='yes',
                                      help='automatically answers yes on prompts'))]
    def execute(self, options, args):
        super(DropDB, self).execute(options, args)
        return self.sys_control.drop_database(yes=options.yes)


class CreateDB(ProductionCommand):
    """Creates the database."""
    keyword = 'createdb'
    def execute(self, options, args):
        super(CreateDB, self).execute(options, args)
        return self.sys_control.create_database()


class BackupDB(ProductionCommand):
    """Backs up the database."""
    keyword = 'backupdb'
    options = ProductionCommand.options +\
              [('-d', '--directory', dict(dest='directory', default='/tmp', help='the direcotry to back up to'))]
    def execute(self, options, args):
        super(BackupDB, self).execute(options, args)
        return self.sys_control.backup_database(options.directory)


class RestoreDB(ProductionCommand):
    """Restores up the database."""
    keyword = 'restoredb'
    options = ProductionCommand.options +\
              [('-f', '--filename', dict(dest='filename', default='/tmp/data.pgsql', help='the file to restore from'))]
    def execute(self, options, args):
        super(RestoreDB, self).execute(options, args)
        return self.sys_control.restore_database(options.filename)


class BackupAllDB(ProductionCommand):
    """Backs up all the databases on the host this project config points to."""
    keyword = 'backupall'
    options = ProductionCommand.options +\
              [('-d', '--directory', dict(dest='directory', default='/tmp', help='the direcotry to back up to'))]
    def execute(self, options, args):
        super(BackupAllDB, self).execute(options, args)
        return self.sys_control.backup_all_databases(options.directory)


class RestoreAllDB(ProductionCommand):
    """Restores all the databases on the host this project config points to."""
    keyword = 'restoreall'
    options = ProductionCommand.options +\
              [('-f', '--filename', dict(dest='filename', default='/tmp/data.sql', help='the file to restore from'))]
    def execute(self, options, args):
        super(RestoreAllDB, self).execute(options, args)
        return self.sys_control.restore_all_databases(options.filename)


class SizeDB(ProductionCommand):
    """Prints the current size of the database."""
    keyword = 'sizedb'
    def execute(self, options, args):
        super(SizeDB, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                print 'Database size: %s' % self.sys_control.size_database()
        return 0


class CreateDBTables(ProductionCommand):
    """Creates all necessary tables in the database."""
    keyword = 'createdbtables'
    def execute(self, options, args):
        super(CreateDBTables, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                return self.sys_control.create_db_tables()


class DropDBTables(ProductionCommand):
    """Drops all necessary tables in the database."""
    keyword = 'dropdbtables'
    def execute(self, options, args):
        super(DropDBTables, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                return self.sys_control.drop_db_tables()


class MigrateDB(ProductionCommand):
    """Runs all necessary database migrations."""
    keyword = 'migratedb'
    def execute(self, options, args):
        super(MigrateDB, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                return self.sys_control.migrate_db()


class ListDependencies(ProductionCommand):
    """List all dependency eggs in dependency order."""
    keyword = 'listdeps'
    options = ProductionCommand.options +\
              [('-v', '--verbose', dict(action='store_true', dest='verbose', help='list direct dependencies too'))]
    def execute(self, options, args):
        super(ListDependencies, self).execute(options, args)
        with self.context:
            distributions = ReahlEgg.compute_ordered_dependent_distributions(self.config.reahlsystem.root_egg)
            for distribution in distributions:
                deps = ''
                if options.verbose:
                    deps = u'[%s]' % (u' | '.join([unicode(i) for i in distribution.requires()]))
                print u'%s %s' % (distribution, deps)
        return 0


class RunJobs(ProductionCommand):
    """Runs all registered scripts."""
    keyword = 'runjobs'
    def execute(self, options, args):
        super(RunJobs, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                self.sys_control.do_daily_maintenance()
        return 0



class ProductionCommandline(ReahlCommandline):
    """The main class for invoking commands on projects in production environments."""
    args_re = ''
    def __init__(self, options):
        super(ProductionCommandline, self).__init__(options, ProdShellConfig())




# Copyright 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from threading import Timer
import datetime

import babel.dates 

from nose.tools import istest
from reahl.tofu import  test, Fixture
from reahl.stubble import stubclass, InitMonitor
from reahl.tofu import vassert

from reahl.component.context import ExecutionContext
from reahl.component.i18n import Translator, SystemWideTranslator
from reahl.component.config import Configuration, ConfigSetting

@stubclass(ExecutionContext)
class LocaleContextStub(ExecutionContext):
    test_locale = u'af'
    @property
    def interface_locale(self):
        return self.test_locale

@istest
class I18nTests(object):
    @test(Fixture)
    def basic_usage(self, fixture):
        """A Translator for a particular component can translate messages for that component into
           the interface_locale on its current ExecutionContext."""
        
            
        _ = Translator(u'reahl-component')  # Will find its translations in the compiled messages of reahl-component

        with LocaleContextStub() as context:
            context.test_locale = u'en_gb'
            vassert( _(u'test string') == u'test string' )
            vassert( _.gettext(u'test string') == u'test string' )
            vassert( _.ngettext(u'thing', u'things', 1) == u'thing' )
            vassert( _.ngettext(u'thing', u'things', 3) == u'things' )

            context.test_locale = u'af'
            vassert( _(u'test string') == u'toets string' )
            vassert( _.gettext(u'test string') == u'toets string' )
            vassert( _.ngettext(u'thing', u'things', 1) == u'ding' )
            vassert( _.ngettext(u'thing', u'things', 3) == u'goeters' )

    @test(Fixture)
    def formatting(self, fixture):
        """A Translator can be used to easily obtain the current locale for use
           by other i18n tools."""
            
        _ = Translator(u'reahl-component') 

        date = datetime.date(2012, 1, 10)

        with LocaleContextStub() as context:
            context.test_locale = u'en_gb'
            vassert( _.current_locale == u'en_gb' )
            actual = babel.dates.format_date(date, format=u'long', locale=_.current_locale)
            vassert( actual == u'10 January 2012' )

            context.test_locale = u'af'
            vassert( _.current_locale == u'af' )
            actual = babel.dates.format_date(date, format=u'long', locale=_.current_locale)
            vassert( actual == u'10 Januarie 2012' )


    @test(Fixture)
    def translator_singleton(self, fixture):
        """Only one SystemwideTranslator is ever present per process."""

        SystemWideTranslator.instance = None  # To "reset" the singleton, else its __init__ will NEVER be called in this test

        with InitMonitor(SystemWideTranslator) as monitor:
            _ = Translator(u'reahl-component')
            _2 = Translator(u'reahl-component')
            
            _(u'test string')
            _.ngettext(u'thing', u'things', 1) 
            _.ngettext(u'thing', u'things', 2) 

            _2(u'test string')
            
        vassert( monitor.times_called == 1 )

    @test(Fixture)
    def translator_singleton_thread_safety(self, fixture):
        """The SystemwideTranslator.get_instance() is this thread-safe."""
        SystemWideTranslator.instance = None  # To "reset" the singleton, else its __init__ will NEVER be called in this test
        SystemWideTranslator.get_instance().map_lock.acquire()
        self.lock_released = False
        def release_lock():
            SystemWideTranslator.get_instance().map_lock.release()
            self.lock_released = True
        timer = Timer(.1, release_lock)
        timer.start()
        _ = Translator(u'reahl-component')
        _.gettext(u'test string')
        timer.cancel()
        vassert( self.lock_released )

    @test(Fixture)
    def translated_config(self, fixture):
        """Configuration can be translated by adding a duplicate setting for each additional locale."""

        class MyConfig(Configuration):
            setting = ConfigSetting(default=u'the default')
            
        config = MyConfig()
        
        with LocaleContextStub() as context:
            context.test_locale = u'en_gb'
            vassert( config.setting == u'the default' )

            context.test_locale = u'af'
            vassert( config.setting == u'the default' )

            config.setting_af = u'die verstek'
            vassert( config.setting == u'die verstek' )



# Copyright 2007-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Classes that aid in dealing with Eggs and setting them up."""

import os
import os.path
from itertools import chain
import logging

from setuptools import setup
from pkg_resources import require, working_set, resource_filename, Requirement, Distribution, iter_entry_points, resource_listdir, resource_isdir

from reahl.component.decorators import memoized

class Attachment(object):
    def __init__(self, filename, label):
        self.filename = filename
        self.label = label

class ReahlEgg(object):
    interface_cache = {}

    def __init__(self, distribution):
        self.distribution = distribution

    @property
    def name(self):
        return self.distribution.key

    @property
    def version(self):
        return self.distribution.version

    @property
    def configuration_spec(self):
        entry_point_dict = self.distribution.get_entry_map().get(u'reahl.configspec', {})
        entry_point = entry_point_dict.get(u'config', None)
        return entry_point.load() if entry_point else None

    def read_config(self, config):
        if self.configuration_spec:
            config.read(self.configuration_spec)

    def validate_config(self, config):
        if self.configuration_spec:
            config.validate_required(self.configuration_spec)

    def list_config(self, config):
        if self.configuration_spec:
            return config.list_required(self.configuration_spec)
        return []

    def get_persisted_classes_in_order(self, orm_control):
        return self.get_ordered_classes_exported_on('reahl.persistlist')
        
    def find_attachments(self, label):
        return [Attachment(i, label) for i in self.get_ordered_names_exported_on('reahl.attachments.%s' % label)]

    @property
    def migrations_in_order(self):
        return self.get_ordered_classes_exported_on('reahl.migratelist')

    def compute_migrations(self, current_schema_version):
        return [cls() for cls in self.migrations_in_order
                      if cls.is_applicable(current_schema_version, self.version)]

    def get_ordered_names_exported_on(self, entry_point):
        entry_point_dict = self.distribution.get_entry_map().get(entry_point, {})
        ordered_names = [(int(name.split(':')[0]), ':'.join(name.split(':')[1:])) for name in entry_point_dict.keys()]
        return [name for order, name in sorted(ordered_names)]

    def get_ordered_classes_exported_on(self, entry_point):
        entry_point_dict = self.distribution.get_entry_map().get(entry_point, {})
        return [entry.load() for order, entry in sorted([(int(order), e) for order, e in entry_point_dict.items()])]

    def get_classes_exported_on(self, entry_point):
        entry_point_dict = self.distribution.get_entry_map().get(entry_point, {})
        return [entry.load() for name, entry in entry_point_dict.items()]

    @property
    def translation_pot_filename(self):
        entry_point_dict = self.distribution.get_entry_map().get(u'reahl.translations')
        translations_package_name = entry_point_dict[self.name].module_name
        translations_file_path = translations_package_name.replace(u'.', u'/')
        return self.distribution.get_resource_filename(self.distribution, '%s/%s' % (translations_file_path, self.name))

    @property
    def scheduled_jobs(self):
        return self.get_classes_exported_on('reahl.scheduled_jobs')
    
    def do_daily_maintenance(self):
        for job in self.scheduled_jobs:
            job()

    def get_roles_to_add(self):
        "returns a list of role classes"
        return []

    @property
    def locale_dirname(self):
        return self.distribution.get_resource_filename(working_set, u'i18n')

    @classmethod
    def get_egg_internal_path_for(cls, translations_entry_point):
        module = translations_entry_point.load()
        dir_or_egg_name = translations_entry_point.dist.location.split(os.sep)[-1]
        paths = [p for p in module.__path__ if p.find(u'%s%s' % (dir_or_egg_name, os.path.sep)) > 0]
        unique_paths = set([p.split(u'%s/' % dir_or_egg_name)[-1] for p in paths])
        assert len(unique_paths) <=1, \
            u'Only one translations package per component is allowed, found %s for %s' % (paths, translations_entry_point.dist)
        assert len(unique_paths) >0, \
            u'No translations found for %s, did you specify a translations package and forget to add locales in there?' % translations_entry_point.dist
        return unique_paths.pop()

    @classmethod
    @memoized
    def get_languages_supported_by_all(cls, root_egg):
        egg_interfaces = cls.get_all_relevant_interfaces(root_egg)
        default_languages = [u'en_gb']
        if not egg_interfaces:
            return default_languages

        domains_in_use = [e.name for e in egg_interfaces]

        languages_for_eggs = {}
        for translation_entry_point in iter_entry_points(u'reahl.translations'):
            requirement = translation_entry_point.dist.as_requirement()
            egg_internal_path = cls.get_egg_internal_path_for(translation_entry_point)

            if resource_isdir(requirement, egg_internal_path):
                languages = [d for d in resource_listdir(requirement, egg_internal_path)
                             if resource_isdir(requirement, '%s/%s' % (egg_internal_path, d))]
            else:
                logging.error('Translations of %s not found in %s' % (requirement, egg_internal_path))
                languages = []

            for language in languages:
                language_path = '%s/%s/LC_MESSAGES' % (egg_internal_path, language)
                domains = [d[:-3] for d in resource_listdir(requirement, language_path) if d.endswith(u'.mo')]
                for domain in domains:
                    if domain in domains_in_use:
                        languages = languages_for_eggs.setdefault(domain, set())
                        languages.add(language)

        if not languages_for_eggs.values():
            return default_languages
        languages = (list(languages_for_eggs.values()))[0].intersection(*languages_for_eggs.values())
        languages.update(default_languages)
        return list(languages)

    @classmethod
    def do_daily_maintenance_for_egg(cls, root_egg):
        eggs_in_order = ReahlEgg.get_all_relevant_interfaces(root_egg)
        for egg in eggs_in_order:
            if isinstance(egg, ReahlEgg):
                egg.do_daily_maintenance()

    @classmethod
    def clear_cache(cls):
        cls.interface_cache.clear()

    @classmethod
    def get_all_relevant_interfaces(cls, main_egg):
        # (We cache the result of this computation, since it is
        #  quite expensive and is called often in the tests.)
        interfaces = cls.interface_cache.get(main_egg, None)
        if interfaces:
            return interfaces
        interfaces = cls.compute_all_relevant_interfaces(main_egg)
        cls.interface_cache[main_egg] = interfaces
        return interfaces

    @classmethod
    def topological_sort(cls, distributions):
    # Algorithm from: http://www.logarithmic.net/pfh-files/blog/01208083168/sort.py
    # See also: http://en.wikipedia.org/wiki/Topological_sorting
        
        graph = {}
        for dist in distributions:
            dependencies = [working_set.find(i) for i in dist.requires()]
            my_requirements =  dist.requires()
            #we want the subset of stuff in the basket we actually depend on, not just the basket itself
            basket_requirements = [i for i in my_requirements
                                   if i.extras]
            for basket in basket_requirements:
                dependencies.extend([working_set.find(Requirement.parse(i)) for i in basket.extras])
                
            graph[dist] = dependencies
        
        count = { }
        for node in graph:
              count[node] = 0
        for node in graph:
            for successor in graph[node]:
                count[successor] += 1

        ready = [ node for node in graph if count[node] == 0 ]
        
        result = []
        while ready:
            node = ready.pop(-1)
            result.append(node)
            
            for successor in graph[node]:
                count[successor] -= 1
                if count[successor] == 0:
                    ready.append(successor)
        
        assert set(distributions) == set(result)
        return result

    @classmethod 
    def get_eggs_for(cls, main_egg):
        distributions = require(str(main_egg))
        return list(set(distributions)) # To get rid of duplicates

    @classmethod
    def compute_ordered_dependent_distributions(cls, main_egg):
        return cls.topological_sort(cls.get_eggs_for(main_egg))
    
    @classmethod
    def compute_all_relevant_interfaces(cls, main_egg):
        interfaces = []

        for i in cls.compute_ordered_dependent_distributions(main_egg):
            entry_map = i.get_entry_map(u'reahl.eggs')
            if entry_map:
                classes = entry_map.values()
                assert len(classes) == 1, u'Only one eggdeb class per egg allowed'
                interfaces.append(classes[0].load()(i))


        return interfaces




# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from nose.tools import istest
from reahl.tofu import  test, Fixture
from reahl.tofu import vassert

from reahl.component.context import ExecutionContext
from reahl.stubble import EmptyStub

@istest
class ContextTests(object):
    @test(Fixture)
    def execution_context_basics(self, fixture):
        """An ExecutionContext is like a global variable for a particular call stack. To create an
           ExecutionContext for a call stack, use it in a with statement."""

        def do_something():
            fixture.found_context = ExecutionContext.get_context()
        def do_high_level_something():
            do_something()
            
        fixture.found_context = None
        some_context = ExecutionContext()
        with some_context:
            do_high_level_something()

        vassert( fixture.found_context is some_context )
        
    @test(Fixture)
    def execution_context_stacking(self, fixture):
        """When an ExecutionContext overrides a deeper one on the call stack, it will retain the same id."""
        some_context = ExecutionContext()

        vassert( some_context is not fixture.context )
        vassert( some_context.id == fixture.context.id )

        with some_context:
            vassert( ExecutionContext.get_context_id() == some_context.id )

    @test(Fixture)
    def contents(self, fixture):
        """A Session, Config or SystemControl may be set on the ExecutionContext."""
        some_context = ExecutionContext()

        session = EmptyStub()
        config = EmptyStub()
        system_control = EmptyStub()

        some_context.set_session( session )
        some_context.set_config( config )
        some_context.set_system_control( system_control )

        vassert( some_context.session is session )
        vassert( some_context.config is config )
        vassert( some_context.system_control is system_control )







# Copyright 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from contextlib import contextmanager

from reahl.tofu import Fixture, test, vassert

from reahl.component.dbutils import ORMControl
from reahl.component.eggs import ReahlEgg
from reahl.component.migration import Migration

# current schema version is saved when tables are creates
# when you run upgrade, only the relevant migrations are run, in order - 2 phases, then new schema version is notes



class ReahlEggStub(ReahlEgg):
    def __init__(self, name, version, migrations):
        super(ReahlEggStub, self).__init__(None)
        self._name = name
        self._version = version
        self.migrations = migrations
    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def migrations_in_order(self):
        return self.migrations

class ORMControlStub(ORMControl):
    def __init__(self):
        self.versions = {}
    @contextmanager
    def managed_transaction(self):
        yield
    def update_schema_version_for(self, egg):
        self.versions[egg.name] = egg.version
    def schema_version_for(self, egg):
        return self.versions[egg.name]
    def initialise_schema_version_for(self, egg):
        self.versions[egg.name] = egg.version


class MigrateFixture(Fixture):
    def new_orm_control(self):
        return ORMControlStub()



@test(MigrateFixture)
def upgrading(fixture):

    fixture.ran = []
    class TestMigration(Migration):
        def upgrade(self):
            fixture.ran.append((self.__class__, u'upgrade'))
        def upgrade_cleanup(self):
            fixture.ran.append((self.__class__, u'cleanup'))
    
    class MigrateOneA(TestMigration):
        version = u'0.1'
    class MigrateOneB(TestMigration):
        version = u'0.1'
    class MigrateTwoA(TestMigration):
        version = u'0.1'
    class MigrateTwoB(TestMigration):
        version = u'0.1'
    
    eggs = [ReahlEggStub(u'one', u'0.0', [MigrateOneA, MigrateOneB]), 
            ReahlEggStub(u'two', u'0.0', [MigrateTwoA, MigrateTwoB])]
    for egg in eggs:
        fixture.orm_control.initialise_schema_version_for(egg)
    for egg in eggs:
        egg._version = u'0.1'
    fixture.orm_control.migrate_db(eggs)

    expected_migrations = [(MigrateTwoA, u'upgrade'), (MigrateTwoB, u'upgrade'), 
                           (MigrateOneA, u'upgrade'), (MigrateOneB, u'upgrade'), 
                           (MigrateOneA, u'cleanup'), (MigrateOneB, u'cleanup'), 
                           (MigrateTwoA, u'cleanup'), (MigrateTwoB, u'cleanup')]
    vassert ( fixture.ran == expected_migrations )

    #case: running it again does not trigger running the upgrade again if the version did not change
    fixture.ran = []
    fixture.orm_control.migrate_db(eggs)
    vassert( fixture.ran == [] )







# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pkg_resources

from nose.tools import istest
from reahl.tofu import  test, Fixture
from reahl.tofu  import vassert, expected, NoException
from reahl.stubble import easter_egg, EmptyStub

from reahl.component.eggs import ReahlEgg


@istest
class ComponentTests(object):
    @test(Fixture)
    def flattened_tree_of_eggs(self, fixture):
        """A Reahl application consists of a root egg and all its egg dependencies - with all such components
           often regarded in flattened order of dependence."""
        easter_egg.clear()
        easter_egg.add_dependency(u'reahl-component')
        
        # All eggs for a root egg can be found in dependency order
        components_in_order = ReahlEgg.compute_ordered_dependent_distributions(easter_egg.as_requirement_string())
        component_names_in_order = [i.project_name for i in components_in_order]
        vassert( component_names_in_order == [easter_egg.project_name, u'reahl-component', u'decorator', u'python-dateutil', u'Babel'] )

    @test(Fixture)
    def interface_with_meta_info(self, fixture):
        """A Reahl component can publish a ReahlEgg instance to supply extra meta information about itself.
           Such interfaces with extra information are also often used from a flattened list in dependency order."""
        
        easter_egg.clear()
        easter_egg.add_dependency(u'reahl-component')
        
        # The interface for a component is published via the reahl.eggs entry point
        line = 'Egg = reahl.component.eggs:ReahlEgg'
        easter_egg.add_entry_point_from_line(u'reahl.eggs', line)
        
        # Interfaces can be queried in dependency order too
        interfaces_in_order = ReahlEgg.compute_all_relevant_interfaces(easter_egg.as_requirement_string())
        vassert( len(interfaces_in_order) == 2 )  # That of reahl-component itself, and of the easteregg
        [interface] = [i for i in interfaces_in_order if i.distribution is easter_egg]

        # The meta-info that can be obtained via such an interface
        vassert( interface.configuration_spec is None )

        orm_control = EmptyStub()
        vassert( interface.get_persisted_classes_in_order(orm_control) == [] )
        vassert( interface.migrations_in_order == [] )
        vassert( interface.get_roles_to_add() == [] )

        # Hooks for allowing a component to do its own housekeeping
        with expected(NoException):
            interface.do_daily_maintenance() 
















from setuptools import setup
setup(
    name='reahl-component',
    version=u'2.1.0',
    description='The component framework of Reahl.',
    long_description=u"Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nReahl-component is the component that contains Reahl's component framework.\n\nSee http://www.reahl.org/docs/current/tutorial/gettingstarted.d.html for installation instructions. ",
    url=u'http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.messages', 'reahl.component_dev', 'reahl.component'],
    py_modules=['setup'],
    include_package_data=False,
    package_data={'': [u'*/LC_MESSAGES/*.mo']},
    namespace_packages=['reahl', 'reahl.messages'],
    install_requires=[u'Babel>=0.9,<0.10', u'python-dateutil>=1.5,<1.6', u'decorator>=3.4,<3.4'],
    setup_requires=[],
    tests_require=[u'reahl-tofu>=2.1,<2.2', u'reahl-stubble>=2.1,<2.2'],
    test_suite=u'reahl.component_dev',
    entry_points={
        'reahl.translations': [
            u'reahl-component = reahl.messages'    ],
        u'reahl.component.databasecontrols': [
            u'NullDatabaseControl = reahl.component.dbutils:NullDatabaseControl'    ],
        'reahl.eggs': [
            u'Egg = reahl.component.eggs:ReahlEgg'    ],
        'console_scripts': [
            u'reahl-control = reahl.component.prodshell:ProductionCommandline.execute_one'    ],
        u'reahl.component.prodcommands': [
            u'CreateDBUser = reahl.component.prodshell:CreateDBUser',
            u'DropDBUser = reahl.component.prodshell:DropDBUser',
            u'CreateDB = reahl.component.prodshell:CreateDB',
            u'DropDB = reahl.component.prodshell:DropDB',
            u'BackupDB = reahl.component.prodshell:BackupDB',
            u'RestoreDB = reahl.component.prodshell:RestoreDB',
            u'BackupAllDB = reahl.component.prodshell:BackupAllDB',
            u'RestoreAllDB = reahl.component.prodshell:RestoreAllDB',
            u'SizeDB = reahl.component.prodshell:SizeDB',
            u'RunJobs = reahl.component.prodshell:RunJobs',
            u'CreateDBTables = reahl.component.prodshell:CreateDBTables',
            u'DropDBTables = reahl.component.prodshell:DropDBTables',
            u'MigrateDB = reahl.component.prodshell:MigrateDB',
            u'ListConfig = reahl.component.prodshell:ListConfig',
            u'CheckConfig = reahl.component.prodshell:CheckConfig',
            u'ListDependencies = reahl.component.prodshell:ListDependencies'    ],
                 },
    extras_require={}
)



# Copyright 2008-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The Reahl configuration utilities."""

import sys
import os.path
import logging
import tempfile
from logging import config
from contextlib import contextmanager

from pkg_resources import require, iter_entry_points, DistributionNotFound

from reahl.component.eggs import ReahlEgg
from reahl.component.context import ExecutionContext


class ConfigurationException(Exception):
    pass

class ExplicitSettingRequired(object):
    pass

class ConfigSetting(object):
    """Used to define one configuration setting on a :class:`Configuration`.
    
    
       :param default: The default value to be used if not specified in a config file.
       :param description: A user readable description explaining what this ConfigSetting is for.
       :param dangerous: Set this to True, if a warning should be emitted when used with the supplied default value.
       :param automatic: Set this to True for a ConfigSetting which is meant to be used for dependency injection.
    """
    _name = None
    def __init__(self, default=ExplicitSettingRequired, description=u'No description supplied', dangerous=False, automatic=False):
        self.description = description
        self.default = default
        self.dangerous = dangerous
        self.automatic = automatic

    @property
    def defaulted(self):
        return self.default is not ExplicitSettingRequired

    def is_localised(self, obj):
        name = self.name(type(obj))
        for key in dir(obj):
            if key.startswith(u'%s_' % name):
                return True
        return False
        
    def __get__(self, obj, objtype):
        setting_name = self.name(type(obj))
        if self.is_localised(obj):
            context = ExecutionContext.get_context()
            locale = context.interface_locale if context else u''
            try:
                return obj.__dict__[u'%s_%s' % (setting_name, locale)]
            except KeyError:
                pass

        if self.is_set(obj):
            return obj.__dict__[setting_name]

        if self.defaulted:
            return self.default

        raise ConfigurationException('%s was not set' % setting_name)

    def __set__(self, obj, value):
        name = self.name(type(obj))
        obj.__dict__[name] = value

    def name(self, objtype):
        if self._name:
            return self._name
        for cls in objtype.mro():
            for name, value in cls.__dict__.items():
                if value is self:
                    self._name = name
                    return name
        raise AssertionError('Could not deduce name for descriptor %s (%s) %s' % (self, self.description, self.default))

    def is_set(self, obj):
        name = self.name(type(obj))
        return obj.__dict__.has_key(name)

    def is_valid(self, obj):        return self.automatic or self.defaulted or self.is_set(obj)

    def is_dangerous(self, obj):
        return self.dangerous and not self.is_set(obj)


class EntryPointClassList(ConfigSetting):
    """A :class:`ConfigSetting` which is not set by a user at all -- rather, its value (a list of classes
       or other importable Python objects) is read from the entry point named `name`.

       :param name: The name of the entry point to read.
       :param description: (See :class:`ConfigSetting`)
    """
    def __init__(self, name, description=u'Description not supplied'):
        super(EntryPointClassList, self).__init__(default=[], description=description)
        self.name = name
        
    def __get__(self, instance, owner):
        classes = []
        for i in iter_entry_points(self.name):
            try:
                classes.append(i.load())
            except ImportError, e:
                print >> sys.stderr, '\nWARNING: Cannot import %s, from %s' % (i, i.dist)
                print >> sys.stderr, e
            except DistributionNotFound, e:
                print >> sys.stderr, '\nWARNING: Cannot find %s, required by %s' % (e, i.dist)
                
        return classes


class MissingValue(object):
    def __repr__(self):
        return _(u'MISSING!!')

    
class Configuration(object):
    """A collection of ConfigSettings for a component. To supply configuration for your component,
       subclass from this class and assign each wanted ConfigSetting as a class attribute. Assign
       the required `filename` and `config_key` class attributes in your subclass as well. The resultant
       class should also be listed in the .reahlproject file of your component in a <configuration> element.
    """
    
    filename = None   #: The name of the config file from which this Configuration will be read.
    config_key = None #: The variable name to which an instance of this Configuration will be bound when reading `filename`
    
    def update(self, other):
        for name, value in other.__dict__.items():
            setattr(self, name, value)

    def get_from_string(self, string_key):
        return self.composite_get_attr(string_key.split('.'))

    def composite_get_attr(self, composite_keys):
        config = self
        for key in composite_keys:
            config = getattr(config, key)
        return config

    def get_config_item(self, name):
        return self.__class__.__dict__[name]

    def has_config_item(self, name):
        return isinstance(self.__class__.__dict__.get(name, None), ConfigSetting)
        
    def config_items(self):
        return [(i, self.get_config_item(i))
                for i in dir(self)
                if self.has_config_item(i)]
    
    def validate_contents(self, filename, composite_key, strict_validation):
        unspecifieds = ['%s.%s' % (composite_key, name)
                        for name, config_item in self.config_items()
                        if not config_item.is_valid(self)]
        if unspecifieds:
            raise ConfigurationException('the following configuration parameters were expected in %s, but not found: %s' % \
                                             (filename, ', '.join(unspecifieds)))

        for name, config_item in self.config_items():
            composite_name = '%s.%s' % (composite_key, name)
            if config_item.is_dangerous(self):
                if strict_validation:
                    raise ConfigurationException('%s in %s is using a dangerous default setting: %s' % (composite_name, filename, config_item.default))
                else:
                    logging.getLogger(__name__).warning('%s in %s is using a dangerous default setting: %s' % (composite_name, filename, config_item.default))

    def list_contents(self, filename, composite_key):
        contents = []
        for name, config_item in self.config_items():
            value = MissingValue()
            try:
                value = getattr(self, name)
            except ConfigurationException:
                pass
            composite_name = '%s.%s' % (composite_key, name)
            contents.append( (filename, composite_name, value, config_item) )
        return contents
                                             
    def configure(self, validate=True):
        pass

    def do_injections(self, config):
        pass
        

class NullORMControl(object):
    def do_nothing(self, *args, **kwargs):
        pass

    def __getattr__(self, name):
        return self.do_nothing

    @contextmanager
    def managed_transaction(self):
        yield None


class ReahlSystemConfig(Configuration):
    filename = u'reahl.config.py'
    config_key = u'reahlsystem'
    root_egg = ConfigSetting(description=u'The root egg of the project', default=os.path.basename(os.getcwd()), dangerous=True)
    connection_uri = ConfigSetting(description=u'The database connection URI', default=u'sqlite:///%s' % os.path.join(tempfile.gettempdir(), u'reahl.db'), dangerous=True)
    orm_control = ConfigSetting(default=NullORMControl(), description=u'The ORM control object to be used', automatic=True)
    debug = ConfigSetting(default=False, description=u'Enables more verbose logging')
    databasecontrols = EntryPointClassList(u'reahl.component.databasecontrols', description=u'All available DatabaseControl classes')
    translation_packages = EntryPointClassList(u'reahl.translations', description=u'All available packages containing translation messages')
    serialise_parallel_requests = ConfigSetting(default=False, description=u'Whether concurrent requests to the web application should be forcibly serialised')


class ConfigAsDict(dict):
    def __init__(self, config):
        self.config = config
        self['config'] = config

    def __getitem__(self, key):
        if self.has_key(key):
            return dict.__getitem__(self,key)
        if hasattr(self.config, key):
            return getattr(self.config, key)
        raise KeyError(key)

    def update_required(self, key):
        try:
            new_value = dict.__getitem__(self,key)
            setattr(self.config, key, new_value)
        except KeyError:
            pass


class StoredConfiguration(Configuration):
    def __init__(self, config_directory_name, strict_validation=False):
        self.config_directory = config_directory_name
        self.strict_validation = strict_validation

    def configure(self, validate=True):
        #http://mail.python.org/pipermail/tutor/2005-August/040993.html
        reload(sys); #read setdefaultencoding python docs - it "enables" the method again
        sys.setdefaultencoding('utf-8')
        
        self.configure_logging()
        logging.getLogger(__name__).info('Using config in %s' % self.config_directory)
        sys.path.insert(0,self.config_directory)

        assert os.path.isdir(self.config_directory), \
               'no such directory: %s' % self.config_directory

        self.read(ReahlSystemConfig)
        self.validate_required(ReahlSystemConfig)

        require(self.reahlsystem.root_egg)
        self.configure_components()
        if validate:
            self.validate_components()
#        sys.path.remove(self.config_directory)

    def configure_logging(self):
        logging_config_file = os.path.join(self.config_directory, 'logging.conf')
        if os.path.isfile(logging_config_file):
            config.fileConfig(logging_config_file)
        else:
            logging.basicConfig()

    def configure_components(self):
        eggs = ReahlEgg.get_all_relevant_interfaces(self.reahlsystem.root_egg)
        for egg in reversed(eggs):
            logging.getLogger(__name__).debug('going to read config for %s' % egg)
            egg.read_config(self)

    def validate_components(self):
        eggs = ReahlEgg.get_all_relevant_interfaces(self.reahlsystem.root_egg)
        for egg in reversed(eggs):
            logging.getLogger(__name__).debug('going to validate config for %s' % egg)
            egg.validate_config(self)

    def list_all(self):
        all = []
        all.extend(self.list_required(ReahlSystemConfig))
        eggs = ReahlEgg.get_all_relevant_interfaces(self.reahlsystem.root_egg)
        for egg in reversed(eggs):
            all.extend(egg.list_config(self))
        return all

    def read(self, configuration_class):
        new_config = self.create_config(configuration_class)
        file_path = os.path.join(self.config_directory, new_config.filename)
        if os.path.isfile(file_path):
            locals_dict = ConfigAsDict(self)
            execfile(file_path, globals(), locals_dict)
            locals_dict.update_required(new_config.config_key)
        else:
            message = 'file "%s" not found, using defaults' % file_path
            logging.getLogger(__name__).info(message)

        unconfigured = self.composite_get_attr(new_config.config_key.split('.'))
        unconfigured.configure() 
        unconfigured.do_injections(self) 

    def create_config(self, configuration_class):
        composite_key = configuration_class.config_key
        keys = composite_key.split('.')
        config = self.composite_get_attr(keys[:-1])
        new_config = configuration_class()
        setattr(config, keys[-1], new_config)
        return new_config

    def validate_required(self, configuration_class):
        composite_key = configuration_class.config_key
        filename = configuration_class.filename
        keys = composite_key.split('.')
        full_filename = os.path.join(self.config_directory, filename)

        try:
            src_config = self.composite_get_attr(keys)
        except AttributeError, e:
            # we make sure that exceptions unrelated to our keys are reraised:
            if set(keys) | set(e.args[0].split()):
                raise ConfigurationException('%s not set in %s' % (composite_key, full_filename))
            else:
                raise

        if not isinstance(src_config, configuration_class):
            raise ConfigurationException('%s is not a %s in %s' % (composite_key, configuration_class, full_filename))

        src_config.validate_contents(full_filename, composite_key, self.strict_validation)

    def list_required(self, configuration_class):
        composite_key = configuration_class.config_key
        items = []
        keys = composite_key.split('.')

        src_config = self.composite_get_attr(keys)
        if not isinstance(src_config, configuration_class):
            raise ConfigurationException('%s is not a %s in %s' % (composite_key, configuration_class, configuration_class.filename))

        items.extend( src_config.list_contents(configuration_class.filename, composite_key) )
        return items


class CodedConfiguration(StoredConfiguration, dict):
    def __init__(self):
        super(CodedConfiguration, self).__init__('<in memory>')

    def read(self, configuration_class):
        fimename = configuration_class.filename
        new_config = self.create_config(configuration_class)
        if self.has_key(filename):
            locals_dict = ConfigAsDict(self)
            self[filename](self)
            locals_dict.update_required(new_config.config_key)
        else:
            message = 'key for file "%s" not found, using defaults' % filename
            logging.getLogger(__name__).info(message)

        self.validate_required(configuration_class)





# Copyright 2009, 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A basic framework for writing commandline utilities."""

import sys
import os.path
import logging
import re
import subprocess
import os


from optparse import OptionParser
import shlex


class ExecutableNotInstalledException(Exception):
    def __init__(self, executable_name):
        self.executable_name = executable_name
    def __str__(self):
        return "Executable not found: %s" % self.executable_name


class Executable(object):
    def __init__(self, name):
        self.name = name

    @property
    def executable_file(self):
        return self.which(self.name)
        
    def which(self, program):
        def is_exe(fpath):
            return os.path.exists(fpath) and os.access(fpath, os.X_OK)

        def ext_candidates(fpath):
            for ext in os.environ.get("PATHEXT", "").split(os.pathsep):
                yield fpath + ext
            yield fpath

        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                exe_file = os.path.join(path, program)
                for candidate in ext_candidates(exe_file):
                    if is_exe(candidate):
                        return candidate

        raise ExecutableNotInstalledException(program)

    def call(self, commandline_arguments, *args, **kwargs):
        return subprocess.call([self.executable_file]+commandline_arguments, *args, **kwargs)
    def check_call(self, commandline_arguments, *args, **kwargs):
        return subprocess.check_call([self.executable_file]+commandline_arguments, *args, **kwargs)
    def Popen(self, commandline_arguments, *args, **kwargs):
        return subprocess.Popen([self.executable_file]+commandline_arguments, *args, **kwargs)


class CommandNotFound(Exception):
    pass

class Command(object):
    """This is the superclass of all Commands that a ReahlCommandline can perform.

    It provides a common place where optparse is invoked. New commands are implemented by subclassing
    this one and overriding its execute method. The class attributes can also be overridden to specify
    how the commandline should be parsed.
    """

    options = []
    keyword = 'not implemented'
    usage_args = u''

    def __init__(self, commandline):
        self.commandline = commandline
        self.parser = OptionParser()
        self.parser.set_usage(u'%s %s' % (os.path.basename(sys.argv[0]), u'%s [options] %s' % (self.keyword, self.usage_args)))
        for (short_version, long_version, kwargs) in self.options:
            self.parser.add_option(short_version, long_version, **kwargs)

    def parse_commandline(self, line):
        options, args = self.parser.parse_args(shlex.split(line))
        self.verify_commandline(options, args)
        return options, args

    def do(self, line):
        try:
            options, args = self.parse_commandline(line)
            return self.execute(options, args)
        except SystemExit, ex:
            return ex.code

    def help(self):
        self.parser.print_help()

    def verify_commandline(self, options, args):
        pass

    def execute(self, options, args):
        pass


class ReahlCommandline(object):
    """A generic class for invoking commands on the commandline."""
    command_entry_point = None
    usage_string = '[options] <command> [command options]'
    args_re = '((-l|--loglevel) +[\w\+\-\'"\>]+)?'
    options = [('-l', '--loglevel', dict(dest='loglevel', default='WARNING',
                                         help='set log level to this'))]

    def __init__(self, options, config):
        self.options = options
        self.commands = config.commands

    @property
    def command_names(self):
        return [i.keyword for i in self.commands]

    def command_named(self, name):
        for i in self.commands:
            if i.keyword == name:
                return i(self)
        raise CommandNotFound(name)

    def print_usage(self, parser):
        print >> sys.stderr, parser.format_help()
        print >> sys.stderr, '\nCommands: ( %s <command> --help for usage on a specific command)\n' % os.path.basename(sys.argv[0])
        max_len = max([len(command.keyword) for command in self.commands])
        format_template = '{0: <%s}\t{1}' % max_len
        for command in self.commands:
            print >> sys.stderr, format_template.format(command.keyword, command.__doc__)
        print >> sys.stderr, ''

    @classmethod
    def execute_one(cls):
        """The entry point for running command from the commandline."""

        command, line, options, parser = cls.parse_commandline()
        return_code = None

        if command:
            commandline = cls(options)
            if command in ['-h', '--help']:
                commandline.print_usage(parser)
                return_code = -1
            else:
                return_code = commandline.execute_command(command, line, options, parser)
        return return_code

    def set_log_level(self, log_level):
        loglevel = getattr(logging, log_level)
        logging.getLogger(u'').setLevel(log_level)
        
    def execute_command(self, command, line, options, parser):
        self.set_log_level(options.loglevel)
        if command in self.command_names:
            return self.command_named(command).do(line)
        self.print_usage(parser)
        return None

    @classmethod
    def parse_commandline(cls):
        parser = OptionParser(usage='%s %s' % (os.path.basename(sys.argv[0]), cls.usage_string))
        for (short_version, long_version, kwargs) in cls.options:
            parser.add_option(short_version, long_version, **kwargs)

        args_string = ' '.join(sys.argv[1:])
        args_match = re.match(cls.args_re, args_string)
        if args_match:
            initial_args = shlex.split(args_string[:args_match.end()])
            remaining_args = sys.argv[len(initial_args)+1:]
        else:
            parser.error('Could not parse commandline')

        options, args = parser.parse_args(initial_args)

        if not len(remaining_args) >= 1:
            return ('-h', None, options, parser)

        # Note: its necessary to quote parts before we join them, else " chars are nuked (if there were any)
        try:
            command, line = remaining_args[0], ' '.join(['\'%s\'' % i for i in remaining_args[1:]])
        except IndexError:
            parser.error('No commands specified, try %s --help for usage' % os.path.basename(sys.argv[0]))
        return command, line, options, parser


